/**
 * @method serialize
 * @param value {any}
 * @returns {string}
 */
var serialize = function (value) { return JSON.stringify(value); };
/**
 * @method deserialize
 * @param str {string}
 * @returns {any}
 */
var deserialize = function (str) { return JSON.parse(str); };


var cm = easyar.ComponentManager.getInstance();
var s = easyar.Scene.getInstance();

var arcamera = easyar.ARCameraPrefab.createOnObject(s.rootObject());
var cam_matrix = easyar.Matrix44F.lookAt({ x: 0, y: 1, z: 5 }, { x: 0, y: 1, z: 0 }, { x: 0, y: 1, z: 0 });
cam_matrix.inverse();
arcamera.transform.matrix = cam_matrix;
//arcamera.setCenterMode("augmenter");

var com_cameraDevice;
var com_augmenter;
arcamera.components().forEach(function (i) {
    if (i.kind == "cameraDevice") {
        com_cameraDevice = i;
    }
    if (i.kind == "augmenter") {
        com_augmenter = i;
    }
});
//com_cameraDevice.deviceType = "front";
//com_cameraDevice.horizontalFlip = true;

arcamera.open();

arcamera.children().forEach(function (i) {
    i.components().forEach(function (i) {
        if (i.kind == "camera") {
            s.setMainCamera(i);
        }
    });
});

var imagetracker = easyar.ImageTrackerPrefab.createOnObject(s.rootObject());
imagetracker.bindCamera(arcamera);

var com_imagetracker;
imagetracker.components().forEach(function (i) {
    if (i.kind == "imageTracker") {
        com_imagetracker = i;
    }
});

// choose which tracker to use here after you call preLoadTarget from native interface,
// native callbacks will be triggered afterwards.
s.setPreLoadListener(function (targetDesc) {
    console.log("load target from scene: " + targetDesc);
    imagetracker.preLoadTarget(targetDesc);
});

s.setManifestTargetLoadListener(function (target) {
    target.bindTracker(imagetracker);
});

var targets = [];
var currentTarget = {};
com_imagetracker.addTargetLoadListener(function (tracker, target, status) {
    target.hideOnLost = false;
    var IsContains = false;
    targets.forEach(function (i) {
        if (i != currentTarget) {
            i.object().children().forEach(function (i) {
                i.active = false;
            });
        } else {
            IsContains = true;
        }
    });
    if (IsContains) {
        targets.push(target);
    }
    console.log("global load (" + status + "): " + target.id + " (" + target.targetName + ")");
});
com_imagetracker.addTargetUnloadListener(function (tracker, target, status) {
    console.log("global unload (" + status + "): " + target.id + " (" + target.targetName + ")");
});

com_augmenter.addTargetFoundListener(function (augmenter, target) {
    currentTarget = target;
    JsNative.targetFound(target.uid);
    IMUReset.Reset = false;
    imu.disableIMU();
    console.log("global found: " + target.id + " (" + target.targetName + ")");
});
com_augmenter.addTargetLostListener(function (augmenter, target) {
    JsNative.targetLost(target.uid);
    IMUReset.Reset = true;
    imu.enableIMU();
    console.log("global lost: " + target.id + " (" + target.targetName + ")");
});


/**
  * @param sendMessage {(name:string, params:string[]) => void}
  * @param pushResponseHandler {(name:string, resultReceiver:string => void) => void}
  */
var JsNativeBinding = function (sendMessage, pushResponseHandler) {
    return {
        openWebView: function (url) {
            if (url == undefined) {
                url = "";
            }
            sendMessage("request:JsNativeBinding.openWebView", [url]);
        },
        getCameraDeviceType: function (type) {
            sendMessage("response:JsNativeBinding.getCameraDeviceType", [type.toString()]);
        },
        targetLost: function (type) {
            sendMessage("request:JsNativeBinding.targetLost", [type]);
        },
        targetFound: function (type) {
            sendMessage("request:JsNativeBinding.targetFound", [type]);
        },
        openView: function (type) {
            sendMessage("request:JsNativeBinding.openView", [type]);
        },
        sendMarkMessage: function (type) {
            sendMessage("request:JsNativeBinding.sendMarkMessage", [type]);
        },
        takePhoto: function (type) {
            sendMessage("request:JsNativeBinding.takePhoto", [type]);
        },
        saveJSMessage: function (w,e) {
            sendMessage("request:JsNativeBinding.saveJSMessage", [w,e]);
        },
        openGesture: function (type) {
            sendMessage("request:JsNativeBinding.openGesture", [type]);
        },
        closeGesture: function (type) {
            sendMessage("request:JsNativeBinding.closeGesture", [type]);
        },
        getJSMessage: function (type) {
            sendMessage("request:JsNativeBinding.getJSMessage", [type]);
        },
        showNativeButton: function (type) {
            sendMessage("request:JsNativeBinding.showNativeButton", [type]);
        },
        closeNativeButton: function (type) {
            sendMessage("request:JsNativeBinding.closeNativeButton", [type]);
        }

    };
};

var NativeJsBinding = function () {
    var functions = {};
    functions["request:NativeJsBinding.resetContent"] = function (receiver, params, sendMessage) {
        receiver.resetContent();
    };
    functions["request:NativeJsBinding.getCameraDeviceType"] = function (receiver, params, sendMessage) {
        receiver.getCameraDeviceType();
    };
    functions["request:NativeJsBinding.changeCameraDeviceType"] = function (receiver, params, sendMessage) {
        receiver.changeCameraDeviceType();
    };
    functions["request:NativeJsBinding.hideAR"] = function (receiver, params, sendMessage) {
        receiver.hideAR();
    };
    functions["request:NativeJsBinding.showARWithTrack"] = function (receiver, params, sendMessage) {
        receiver.showARWithTrack();
    };
    functions["request:NativeJsBinding.showARWithIMU"] = function (receiver, params, sendMessage) {
        receiver.showARWithIMU();
    };
    functions["response:NativeJsBinding.getJSMessage"] = function (receiver, params, sendMessage) {
        receiver.getJSMessage(params[0], params[1]);
    };
    functions["response:NativeJsBinding.GestureEnd"] = function (receiver, params, sendMessage) {
        receiver.GestureEnd(params[0]);
    };
    
    functions["request:NativeJsBinding.camera3d"] = function (receiver, params, sendMessage) {
        receiver.camera3d();
    };
    functions["request:NativeJsBinding.camera2d"] = function (receiver, params, sendMessage) {
        receiver.camera2d();
    };
    return {
        /**
         * @param name {string}
         * @returns {(receiver:INativeJsBindingReceiver, params:string[]) => string}
         */
        tryGetFunction: function (name) {
            if (name in functions) {
                return functions[name];
            } else {
                return null;
            }
        }
    };
};

var responseRegistry = {}
var JsNative = new JsNativeBinding(easyar.Rpc.sendMessage, function (name, resultReceiver) {
    var queue = responseRegistry[name];
    if (queue === undefined) {
        queue = new Array();
        responseRegistry[name] = queue;
    }
    queue.push(resultReceiver);
});

var nativeReceiver = {
    resetContent: function () {
        console.log("[Js] request:NativeJsBinding.resetContent : default");
    },
    getCameraDeviceType: function () {
        var cameraType = com_cameraDevice.deviceType;
        JsNative.getCameraDeviceType(cameraType.toString());
        console.log("[Js] request:NativeJsBinding.getCameraDeviceType : default");
    },
    changeCameraDeviceType: function () {
        IMUReset.Reset = true;
        if (com_cameraDevice.deviceType == "normal" || com_cameraDevice.deviceType == "back") {
            arcamera.close();
            com_cameraDevice.deviceType = "front";
            arcamera.open();
        } else if (com_cameraDevice.deviceType == "front") {
            arcamera.close();
            com_cameraDevice.deviceType = "back";
            arcamera.open();

        }
        console.log("[Js] request:NativeJsBinding.changeCameraDeviceType : default");
    },
    hideAR: function () {
        sceneState.hideAR();
        console.log("[Js] request:NativeJsBinding.changeToLost : default");
    },
    showARWithIMU: function () {
        sceneState.showARWithIMU();
        console.log("[Js] request:NativeJsBinding.showARWithIMU : default");
    },
    showARWithTrack: function () {
        sceneState.showARWithTrack();
        console.log("[Js] request:NativeJsBinding.showARWithTrack : default");
    },
    getJSMessage: function (key,value) {
        console.log("[Js] response:NativeJsBinding.getJSMessage : default " + key + ": "+ value);
    },
    GestureEnd: function (type) {
        console.log("[Js] request:NativeJsBinding.GestureEnd : default " + type);
    },
camera3d: function(){
    sceneState.camera3d();
    console.log("[Js] request:NativeJsBinding.camera3d : default");
},
camera2d: function(){
    sceneState.camera2d();
    console.log("[Js] request:NativeJsBinding.camera2d : default");
}
};

var NativeJs = new NativeJsBinding();
easyar.Rpc.setMessageReceiver(function (name, params) {
    var f = NativeJs.tryGetFunction(name);
    if (f != null) {
        f(nativeReceiver, params, easyar.Rpc.sendMessage);
    } else if (name in responseRegistry) {
        var queue = responseRegistry[name];
        var f = queue.shift();
        if (queue.length == 0) {
            delete responseRegistry[name];
        }
        f(params);
    } else {
        throw new Error("InvalidMethodName: " + name);
    }
});

var sceneState = {
    status: "lost", recentTarget: {},//tracking lost IMU
    hideAR: function () {
        this.status = "lost";
        //com_imagetracker.isRunning = true;
        if (this.recentTarget.object == undefined) {
            return;
        }

        this.recentTarget.object().children().forEach(function (i) {
            i.active = false;
        });
    },
    showARWithTrack: function () {
        this.status = "tracking";
        if (currentTarget.object == undefined) {
            return;
        }

        currentTarget.object().children().forEach(function (i) {
            i.active = true;
        });

    },
    showARWithIMU: function () {
        this.status = "IMU";
        if (currentTarget.object == undefined) {
            return;
        }
        this.recentTarget = currentTarget;
        imuTarget = currentTarget.object();
    }
};

//light
/*
var light_rotations = [{ x: 35 * 3.14 / 180, y: 90 * 3.14 / 180, z: 0 }, { x: 35 * 3.14 / 180, y: 150 * 3.14 / 180, z: 0 }, { x: 16 * 3.14 / 180, y: -25 * 3.14 / 180, z: 10 }, { x: -70 * 3.14 / 180, y: 0, z: 0 }];
light_rotations.forEach(function (i) {
    var obj_light = s.rootObject().createChild();
    obj_light["com_light"] = cm.createComponentOnObject("directionalLight", obj_light, {});
    obj_light["com_light"].diffuse = 1;
    obj_light["com_light"].specular = 0;
    obj_light.transform.rotationXYZ = i;
});
*/
var child_light = s.rootObject().createChild();
var amblight_cm = cm.createComponentOnObject("ambientLight", child_light, {});
amblight_cm.ambient = 1;
child_light.transform.matrix = easyar.Matrix44F.TRS({ x:0, y: 0, z: -1 }, { x: 0, y: 0, z: 0 }, { x: 1, y: 1, z: 1 });

//imu

var imuTarget = s.rootObject().createChild();
imuTarget.name = "imuTarget";

function v_m(v, m) {
    var _v = { x: v.x, y: v.y, z: v.z };
    var _m = m.raw();
    _v.x = v.x * _m[0] + v.y * _m[4] + v.z * _m[8];
    _v.y = v.x * _m[1] + v.y * _m[5] + v.z * _m[9];
    _v.z = v.x * _m[2] + v.y * _m[6] + v.z * _m[10];
    return _v;
}

var imu = cm.createComponentOnObject("imu", imuTarget, {});

s.touch.setTapListener(function (x, y) {

});

var vec_d_distance = function (v1, v2) {
    var _v = { x: 0, y: 0, z: 0 };
    _v.x = v1.x - v2.x;
    _v.y = v1.y - v2.y;
    _v.z = v1.z - v2.z;

    var distance = Math.sqrt(_v.x * _v.x + _v.y * _v.y + _v.z * _v.z);
    return distance;
}

function dot(v1, v2) {
    return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
}

function normalize(v) {
    var len = Math.sqrt(dot(v, v));
    return { x: v.x / len, y: v.y / len, z: v.z / len };
}

function cross(v1, v2) {
    return { x: v1.y * v2.z - v2.y * v1.z, y: v1.z * v2.x - v1.x * v2.z, z: v1.x * v2.y - v1.y * v2.x };
}

function multiply(v, num) {
    return { x: v.x * num, y: v.y * num, z: v.z * num };
}

function add(v1, v2) {
    return { x: v1.x + v2.x, y: v1.y + v2.y, z: v1.z + v2.z };
}

var showInCamera = function (obj, d) {
    var v = { x: 0, y: 0, z: -d };
    v = v_m(v, arcamera.transform.matrix);
    var cam_pos = arcamera.transform.position;
    var pos = add(cam_pos, v);
    var cam_project = { x: cam_pos.x, y: 0, z: cam_pos.z };
    var obj_project = { x: pos.x, y: 0, z: pos.z };
    if (cam_project.x != obj_project.x && cam_project.z != obj_project.z) {
        var matrix = easyar.Matrix44F.lookAt(obj_project, cam_project, { x: 0, y: 1, z: 0 });
        obj.transform.matrix = matrix;
    }
    obj.transform.position = pos;
    obj.transform.scale = { x: 1, y: 1, z: 1 };
}

var vec_d_distance = function (v1, v2) {
    var _v = { x: 0, y: 0, z: 0 };
    _v.x = v1.x - v2.x;
    _v.y = v1.y - v2.y;
    _v.z = v1.z - v2.z;

    var distance = Math.sqrt(_v.x * _v.x + _v.y * _v.y + _v.z * _v.z);
    return distance;
}

function setTouchXYWithGroy(_x, _y) {
    var target_cam = arcamera;
    var t = target_cam.transform;

    var cam_r = { x: 1, y: 0, z: 0 };
    var cam_up = { x: 0, y: 1, z: 0 };
    cam_r = v_m(cam_r, t.matrix);
    cam_up = v_m(cam_up, t.matrix);

    var world_up = { x: 0, y: 1, z: 0 };

    var dot_value_r = dot(cam_r, world_up);
    var dot_value_up = dot(cam_up, world_up);
    var right;
    var right_value;
    var forward_value;
    var orientation;

    if (Math.abs(dot_value_r) > Math.abs(dot_value_up)) {

        if (dot_value_r < 0) {
            orientation = "LandspaceRight";
        } else {
            orientation = "LandspaceLeft";
        }
        right_value = -_y;
        forward_value = -_x;
        right = normalize({ x: cam_up.x, y: 0, z: cam_up.z });
    }
    else {

        if (dot_value_up < 0) {
            orientation = "InPortrait";
        } else {
            orientation = "Portrait";
        }
        right = normalize({ x: cam_r.x, y: 0, z: cam_r.z });
        right_value = _x;
        forward_value = -_y;
    }
    return { x: right_value, y: forward_value, _right: right, _orientation: orientation };
}


var d = vec_d_distance(arcamera.transform.position, imuTarget.transform.position);



s.touch.setSingleMoveListener(function (_x, _y) {

    if (sceneState.status == "tracking") {
        return;
    }

    if (imuTarget.transform == undefined) {
        return;
    }

    d = vec_d_distance(arcamera.transform.position, imuTarget.transform.position);

    var translation = setTouchXYWithGroy(_x, _y);


    if (com_cameraDevice.deviceType == "front") {
        if (translation._orientation == "InPortrait" || translation._orientation == "Portrait") {
            translation.x = -translation.x;
        } else {
            translation.y = -translation.y;
        }
    }

    var cam_right = translation._right;
    var cam_forward = cross({ x: 0, y: 1, z: 0 }, translation._right);
    if(com_cameraDevice.deviceType == "front")
    {
        if(translation._orientation == "InPortrait" || translation._orientation == "LandspaceLeft" )
        {
            translation.y = -translation.y;
        }
        translation = add(multiply(cam_right, translation.x / 1000 * d), multiply({x: 0, y: 1, z: 0}, translation.y / 1000 * d));
    }else{
         translation = add(multiply(cam_right, translation.x / 1000 * d), multiply(cam_forward, translation.y / 1000 * d));
    }

    imuTarget.transform.position = add(imuTarget.transform.position, translation);
});



s.touch.setPinchListener(function (_z) {
    if (sceneState.status == "tracking") {
        return;
    }
    d = vec_d_distance(arcamera.transform.position, imuTarget.transform.position);
    _z = -_z * d / 30;
    var v = { x: 0, y: 0, z: -_z };
    v = v_m(v, arcamera.transform.matrix);
    var pos = imuTarget.transform.position;
    pos.x += v.x;
    pos.y += v.y;
    pos.z += v.z;
    imuTarget.transform.position = pos;
});

s.touch.setDoubleMoveListener(function (_x, _y) {
    d = vec_d_distance(arcamera.transform.position, imuTarget.transform.position);
    var translation = setTouchXYWithGroy(_x, _y);
    if (com_cameraDevice.deviceType == "front") {
        switch (translation._orientation) {
            case "Portrait":
                translation.x = -translation.x;
                break;
            case "InPortrait":
                translation.y = -translation.y;
                break;
            case "LandspaceRight":
                translation.y = -translation.y;
                break;
            case "LandspaceLeft":
                translation.x = -translation.x;
                break;
        }
    }

    if (Math.abs(translation.x) > Math.abs(translation.y)) {
        var r = imuTarget.transform.rotationXYZ;
        r.x = 0;
        r.y += translation.x * 3.14 / 180;
        r.z = 0;
        imuTarget.transform.rotationXYZ = r;
    }
    else {
        if (sceneState.status == "tracking") {
            return;
        }
        var world_up = { x: 0, y: translation.y / 1000 * d, z: 0 };
        imuTarget.transform.position = add(imuTarget.transform.position, world_up);
    }
});




nativeReceiver.resetContent = function () {

    if (sceneState.status == "tracking") {
        return;
    }
    var _d;
    if (currentTarget.size != undefined) {
        _d = currentTarget.size.x;
    } else {
        _d = imuTarget.transform.scale.x;
    }
    showInCamera(imuTarget, _d);

    d = vec_d_distance(arcamera.transform.position, imuTarget.transform.position);
};

var IMUReset = function (obj) {
    return {
        Reset: false, frame: 0, time: 5,
        awake: function () { },
        update: function () {
            if (this.Reset) {
                if (imu.isReady) {
                    if (this.frame < this.time) {
                        this.frame++;
                    } else {
                        this.frame = 0;
                        imuTarget.active = true;
                        nativeReceiver.resetContent();
                        this.Reset = false;
                    }
                }
            } else {
                this.frame = 0;
            }
        },
        onEnable: function () { },
        onDisable: function () { },
        onDestroy: function () { }
    };
};
cm.registerComponent("IMUReset", function (obj) { return new IMUReset(obj) });
var IMUReset = cm.createComponentOnObject("IMUReset", imuTarget, {});
imu.enableIMU();

//JsNative.showNativeButton("changeCamera");
//JsNative.closeNativeButton("changeCamera");

//JsNative.openGesture("back");
//JsNative.openView("view1");
/*sceneState.changeToIMU();
sceneState.changeToTracking();*/
/*
* JsCallNative sample:
    JsNative.openWebView("www.easyar.cn");

* NativeCallJs sample:
    var cube = s.rootObject().createChild();
    nativeReceiver.resetContent = function(){
        cube.transform.position = {x:0,y:0,z:0};
    };
*/
